import pytest
import datetime
from django_login_email import models


@pytest.mark.django_db
def test_models_in_django():
    now_t = datetime.datetime.now()
    e = models.EmailLogin.set_email_last_time(email="svtter@163.com", datetime=now_t)
    assert models.EmailLogin.objects.get(email="svtter@163.com")

    now_t2 = datetime.datetime.now() + datetime.timedelta(minutes=10)
    e.set_last_time(now_t2)
    assert e.last_time == now_t2



import datetime
from typing import Any
from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.generic.edit import FormView
from django.views.generic import TemplateView
from django.contrib.auth import get_user_model
from . import forms
from . import email
from . import models

# Create your views here.


class TimeLimit(email.TimeLimit):
    minutes = 10


class MailRecordModelMixin(email.EmailInfoMixin):
    """Here is an example for MailRecord, using django model. You could implement yourself."""

    def set_mail_record(self, mail: str):
        models.EmailLogin.set_email_last_time(mail, datetime.datetime.now(tz=datetime.timezone.utc))

    def reset_mail(self, mail: str):
        # models.EmailLogin.objects.filter(email=mail).delete()
        e = models.EmailLogin.objects.get(email=mail)
        e.last_time = e.last_time - datetime.timedelta(minutes=self.tl.minutes)
        e.save()

    def get_mail_record(self, mail: str) -> email.MailRecord:
        # for easy to change. use a function.
        try:
            e = models.EmailLogin.objects.get(email=mail)
        except models.EmailLogin.DoesNotExist:
            return email.MailRecord(email=mail, last_time=None)
        return email.MailRecord(email=e.email, last_time=e.last_time)


class EmailLoginView(FormView, MailRecordModelMixin):
    template_name = "login_email/login.html"
    form_class = forms.LoginForm
    email_info_class = email.EmailLoginInfo
    tl = TimeLimit()

    def form_valid(self, form):
        """check the email"""
        User = get_user_model()
        try:
            self.send_login_mail(form.cleaned_data["email"])
        except User.DoesNotExist as e:
            # ignore user missing problem.
            print(e, form.cleaned_data["email"])
            return render(self.request, "login_email/success.html", {"form": form})
        except Exception as e:
            # raise Exception(e)
            print(e)
            return render(self.request, "login_email/error.html", {"error": e})
        return render(self.request, "login_email/success.html", {"form": form})


class EmailVerifyView(TemplateView, email.EmailValidateMixin):
    tl = TimeLimit()

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        raise Http404("Invalid Request.")

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        token = request.GET.get("token", None)
        if token is None:
            raise Http404("Invalid Request")
        try:
            self.verify_login_mail(request=request, token_v=token)
        except Exception as e:
            # TODO: log the error
            print(e)
            raise Http404("Invalid Request")
        return redirect(self.get_success_url())

    def get_success_url(self):
        return reverse("home")


class EmailLogoutView(TemplateView, email.EmailLogoutMixin):
    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        self.logout(request=request)
        return redirect("login")


class HomeView(TemplateView):
    template_name = "login_email/home.html"

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if not self.request.user.is_anonymous and self.request.user.is_authenticated:
            return super().get(request, *args, **kwargs)
        else:
            return redirect("login")



# Handle email token
import base64
import json
import urllib.parse
import typing as t

from Crypto.Cipher import AES
from django.conf import settings
import datetime

Token = str
EmailAndSalt = str
Email = str

TokenDict = t.TypedDict("TokenDict", {"email": Email, "expire_time": int, "salt": str})


class TokenGenerator(object):
    def __init__(self, minutes: int) -> None:
        self.minutes = minutes

    def get_expire_time(self) -> int:
        return int((datetime.datetime.now() + datetime.timedelta(minutes=self.minutes)).timestamp())

    def gen(self, email, salt):
        return json.dumps(
            {
                "email": email,
                "expire_time": self.get_expire_time(),
                "salt": salt,
            }
        )


class TokenManager(object):
    salt: str = "randomsalt"
    key: bytes
    generator: TokenGenerator

    def __init__(self, minutes: int) -> None:
        self.key = settings.SECRET_KEY[:32].encode("utf-8")
        self.generator: TokenGenerator = TokenGenerator(minutes)

    def check_token(self, token_uncrypt: str) -> t.Optional[TokenDict]:
        """check salt and expire-time"""
        token_dict: TokenDict = json.loads(token_uncrypt)
        if token_dict["salt"] == self.salt and token_dict["expire_time"] > int(datetime.datetime.now().timestamp()):
            return token_dict
        return None

    def get_mail(self, token_uncrypt: TokenDict) -> Email:
        return token_uncrypt["email"]

    def encrypt(self, plaintext, key):
        cipher = AES.new(key, AES.MODE_EAX)
        nonce = cipher.nonce
        ciphertext, tag = cipher.encrypt_and_digest(plaintext)
        return nonce + ciphertext + tag

    def decrypt(self, ciphertext, key):
        nonce = ciphertext[:16]
        tag = ciphertext[-16:]
        ciphertext = ciphertext[16:-16]
        cipher = AES.new(key, AES.MODE_EAX, nonce)
        plaintext = cipher.decrypt_and_verify(ciphertext, tag)
        return plaintext

    def encrypt_mail(self, email: Email) -> str:
        """email -> token -> base64 -> utf-8 -> urlquote"""
        # add salt.
        content = self.generator.gen(email, self.salt)
        content = content.encode("utf-8")
        return urllib.parse.quote(base64.b64encode(self.encrypt(content, self.key)).decode("utf-8"))

    def decrypt_token(self, token: Token) -> str:
        t = urllib.parse.unquote(token).encode("utf-8")
        print(t)
        t = base64.b64decode(t)
        return self.decrypt(t, self.key).decode("utf-8")



import pytest
import datetime
from django_login_email import models
from django_login_email.email import EmailInfoMixin, MailRecord, TimeLimit


class MyTimeLit(TimeLimit):
    minutes = 10


class Mixin(EmailInfoMixin):
    tl = MyTimeLit()

    def get_mail_record(self, mail: str) -> MailRecord:
        r1 = MailRecord(
            email="svtter@163.com",
            last_time=datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(minutes=10),
        )
        # r2 = MailRecord(email="svtter@163.com", last_time=datetime.datetime.now(tz=datetime.timezone.utc))
        return r1

    def set_mail_record(self, mail: str):
        pass


@pytest.mark.django_db
def test_mixin():
    mail = "svtter@163.com"
    e = Mixin()
    re = e.get_mail_record(mail)
    assert re.last_time
    # assert re.last_time + datetime.timedelta(minutes=10) <= datetime.datetime.now(tz=datetime.timezone.utc)
    assert e.check_could_send(email=mail)


class Mixin2(EmailInfoMixin):
    tl = MyTimeLit()

    def get_mail_record(self, mail: str) -> MailRecord:
        r2 = MailRecord(email="svtter@163.com", last_time=None)
        return r2

    def set_mail_record(self, mail: str):
        pass


def test_none_of_mixin():
    mail = "svtter@163.com"
    e = Mixin2()
    re = e.get_mail_record(mail=mail)
    assert re.last_time is None



from django.db import models

# Create your models here.


class EmailLogin(models.Model):
    last_time = models.DateTimeField(auto_now_add=True, verbose_name="Last login request time")
    email = models.EmailField(verbose_name="Email", unique=True, null=False)

    @classmethod
    def set_email_last_time(cls, email, datetime) -> "EmailLogin":
        obj, _ = cls.objects.get_or_create(email=email)
        obj.set_last_time(datetime)
        return obj

    def set_last_time(self, datetime):
        self.last_time = datetime
        self.save()

    def __str__(self) -> str:
        return f"Email: {self.email}"



from dataclasses import dataclass
import typing as t
import abc
import string
import datetime
from datetime import timezone

from django.core.mail import EmailMessage
from django.contrib.auth import get_user_model

from django.contrib.auth import login, logout
from . import token


class EmailLoginInfo(object):
    subject: str
    message: str
    welcome_text = "Welcome to meterhub! Please click the link below to login.<br>"
    url: str = "http://127.0.0.1:8000/account/verify?token="
    login_message: string.Template = string.Template('Click <a href="$url$token">Link</a>')
    from_email: str

    def __init__(self) -> None:
        self.set_variables()

    def set_variables(self):
        raise NotImplementedError("You must set the subject and from_email.")

    def set_token(self, value):
        self.message = self.welcome_text + self.login_message.substitute(url=self.url, token=value)


class TimeLimit(object):
    minutes: int = 10


@dataclass
class MailRecord(object):
    last_time: t.Optional[datetime.datetime]
    email: str


class MailRecordMixin(abc.ABC):

    @abc.abstractmethod
    def get_mail_record(self, mail: str) -> MailRecord:
        # for easy to change. use a function.
        raise NotImplementedError("You must implement get_mail_record")

    @abc.abstractmethod
    def set_mail_record(self, mail: str):
        raise NotImplementedError("You must implement set_mail_record")


class EmailInfoMixin(MailRecordMixin):
    email_info_class: t.Type[EmailLoginInfo]

    tl: TimeLimit

    def check_user(self, email):
        User = get_user_model()
        u = User.objects.get(email=email)
        return u

    def check_could_send(self, email):
        re = self.get_mail_record(email)
        if (re.last_time is None) or (
            re.last_time + datetime.timedelta(minutes=self.tl.minutes) <= datetime.datetime.now(tz=timezone.utc)
        ):
            return True
        return False

    def send_valid(self, email: str):
        e = self.email_info_class()
        m = token.TokenManager(self.tl.minutes)

        token_v = m.encrypt_mail(email)
        e.set_token(token_v)

        msg = EmailMessage(e.subject, e.message, e.from_email, [email])
        msg.content_subtype = "html"
        msg.send()

    def send_login_mail(self, email: str):
        self.check_user(email)
        if not self.check_could_send(email=email):
            raise Exception(f"Cannot send. Wait {self.tl.minutes} minutes.")

        self.send_valid(email)
        self.set_mail_record(email)


class EmailValidateMixin(object):
    tl: TimeLimit

    def verify_login_mail(self, request, token_v: str):
        m = token.TokenManager(self.tl.minutes)
        emailAndSalt = m.decrypt_token(token=token_v)
        token_d = m.check_token(emailAndSalt)
        if token_d is None:
            raise Exception("Invalid token.")

        User = get_user_model()
        u = User.objects.get(email=m.get_mail(token_d))
        if not u.is_active:
            raise Exception("Inactive user, disallow login.")

        login(request, u)


class EmailLogoutMixin(object):
    def logout(self, request):
        logout(request)



# Generated by Django 5.0.4 on 2024-04-05 07:43

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='EmailLogin',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_time', models.DateTimeField(auto_now_add=True, verbose_name='Last login request time')),
                ('email', models.EmailField(max_length=254, verbose_name='Email')),
            ],
        ),
    ]
